unit uClass;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Graphics, Math, IniFiles, Dialogs, Forms;

const
  MAXX = 500;
  MAXY = 500;
  MINDIST = 10;

  SPAWNRATE=0.05;
  MANACOST=0.001;
  DAMAGE = 0.1;
  HEALPOWER=0.1;

  MONSTERSPEED=5;

  PLAYERSPEED=10;
  PLAYERARMOR=0.5;

  DEMONSPEED=10;
  GOLEMSPEED=8;

type


  TGame = class;

  { TGameObject }

  TGameObject = class
    Game: TGame;
    X, Y: extended;
    Broken: boolean;
    destructor Destroy; override;
    procedure Draw(C: TCanvas); virtual; abstract;
    constructor Create(aGame: TGame; aX, aY: extended);
    procedure Process; virtual;
  end;
  TGameClass = class of TGameObject;

  { TUnit }

  TUnit = class(TGameObject)
    Speed, TargetX, TargetY, HP, Armor: extended;
    procedure Move;
    procedure Hit;
    procedure Die; virtual;
    procedure SetDirTo(ToX, ToY: extended);
    procedure Draw(C: TCanvas); override;
    procedure Process; override;
    constructor Create(aGame: TGame; aX, aY: extended);virtual;
    procedure Heal;
  end;
  TUnitClass = class of TUnit;

  { TMonster }

  TMonster = class(TUnit)
    procedure Draw(C: TCanvas); override;
    constructor Create(aGame: TGame; aX, aY: extended); override;
    procedure Process; override;
  end;

  { TDemon }

  TDemon = class(TUnit)
    procedure Draw(C: TCanvas); override;
    constructor Create(aGame: TGame; aX, aY: extended);override;
    procedure Process; override;
  end;

  { TGolem }

  TGolem = class(TUnit)
    procedure Draw(C: TCanvas); override;
    constructor Create(aGame: TGame; aX, aY: extended);override;
    procedure Process; override;
  end;



  { TObstracle }

  TObstracle = class(TGameObject)
    function CanStep(Who: TUnit; var aX: extended; var aY: extended): boolean; virtual; abstract;
  end;

  { TRectObstracle }

  TRectObstracle = class(TObstracle)
    X2, Y2: extended;
    constructor Create(aGame: TGame; ax1, ay1, ax2, ay2: extended);
    procedure Draw(C: TCanvas); override;
    function CanStep(Who: TUnit; var aX: extended; var aY: extended): boolean; override;
  end;

  TElement = (Fire, Water, Air, Earth);
  TElements = set of TElement;
  TElementColors = array[TElement]of TColor;
  TElementNames = array[TElement]of string;

  TLineStyle = (Scratch, Current, Ready, Active);
  TLine = class;

  { TSource }

  TSource = class(TGameObject)
    Name: string;
    MP:      extended;
    Element: TElement;
    Lines: TList;
    procedure Draw(C: TCanvas); override;
    function UseMana: boolean;
    constructor Create(aGame: TGame; ax, ay, aMP: extended; aElement: TElement);
    destructor Destroy; override;
    procedure Process; override;
    function Connected(Node: TSource): TLine;
  end;

  TExtendedPoint = packed record
    X, Y: Extended;
  end;

  { TLine }

  TLine = class
    Game: TGame;
    Start, Finish: TSource;
    LastAngle: Real;
    Points: array of TExtendedPoint;
    constructor Create(aGame: TGame; aStart: TSource);
    procedure AddPoint(aX, aY: Extended; Angle: Real);
    function Finalize(aFinish: TSource): boolean;
    procedure Draw(C: TCanvas; Style: TLineStyle);
    function OtherNode(S: TSource): TSource;
  end;

  { TSpell }

  TSpell = class(TGameObject)
    Sources: TList;
    Lines: TList;
    constructor Create(aGame: TGame);
    procedure AddLine(L: TLine);
    procedure Draw(C: TCanvas); override;
    procedure Process; override;
    function UseMana: boolean;
    destructor Destroy;override;
  end;

  { TDemoSpell }

  TDemoSpell = class(TSpell)
    procedure Process; override;
    constructor Create(aGame: TGame);
  end;

  { TChannelSpell }

  TChannelSpell = class(TSpell)
    procedure Process; override;
    constructor Create(aGame: TGame; First, Second: TSource);
  end;

  { TSummonSpell }

  TSummonSpell = class(TSpell)
    Summoned: TUnit;
    SummonClass: TUnitClass;
    constructor Create(aGame: TGame; aClass: TUnitClass);
    procedure Activate;
    procedure Process; override;
  end;

  TAreaSpell = class;

  { TSpecialObstracle }

  TSpecialObstracle = class(TObstracle)
    Points: array of TExtendedPoint;
    Caller: TAreaSpell;
    constructor Create(aGame: TGame; aCaller: TAreaSpell);
    destructor Destroy; override;
    function Inside(aX, aY: Extended): boolean;
    procedure AddPoint(aX, aY: Extended);
    function CanStep(Who: TUnit; var aX: extended; var aY: extended): boolean; override;
    procedure Draw(C: TCanvas); override;
    procedure Process; override;
  end;

  { TAreaSpell }

  TAreaSpell = class(TSpell)
    MyArea: TSpecialObstracle;
    constructor Create(aGame: TGame);
    class function PreciseArea: boolean; virtual; abstract;
    procedure Activate;
    destructor Destroy;override;
    function Callback(U: TUnit; aX, aY: Extended): boolean;virtual; abstract;
  end;

  { TBarrier }

  TBarrier = class(TAreaSpell)
    class function PreciseArea: boolean; override;
    function Callback(U: TUnit; aX, aY: Extended): boolean;override;
    procedure Process; override;
  end;


  { THealSpell }

  THealSpell = class(TAreaSpell)
    class function PreciseArea: boolean; override;
    function Callback(U: TUnit; aX, aY: Extended): boolean;override;
    procedure Process; override;
  end;

  { TDamageSpell }

  TDamageSpell = class(TAreaSpell)
    class function PreciseArea: boolean; override;
    function Callback(U: TUnit; aX, aY: Extended): boolean;override;
    procedure Process; override;
  end;


  TPattern = class
    function Recognize(Game: TGame; Line: TLine): TSpell;virtual; abstract;
  end;

  { TCirclePattern }

  TCirclePattern = class(TPattern)
    function Recognize(Game: TGame; Line: TLine): TSpell;override;
  end;

  TLinkCheckResult = (PartialOK, FullOK, IsFail, LastOK);

  { TObjectList }

  TObjectList = class
    Data: array of TObject;
    constructor Create;
    destructor Destroy; override;
    procedure PushObject(S: TObject);
    procedure PopObject;
    function FindSource(S: TObject): boolean;
    function PrevNode(N: Integer): TObject;

    function X(I: Integer): Extended;
    function Y(I: Integer): Extended;
  end;

  { TSequencePattern }

  TSequencePattern = class(TPattern)
    __FirstNode: TSource;
    __FinalLines, __FinalNodes: TObjectList;
    __Index: integer;
    function MaxLen: Integer;virtual;abstract;
    function CanRepeat: boolean;virtual;
    function RecognizePart(CurIndex: Integer; FromNode: TSource; FirstNode: TSource; Exclude, FinalLines, FinalNodes: TObjectList): TSource;
    function CheckLink(Index: Integer; Line: TLine; Node: TSource; FirstNode: TSource; FinalLines, FinalNodes: TObjectList): TLinkCheckResult;virtual;
    function FullRecognize(aLine: TLine; FinalLines, FinalNodes: TObjectList): TSource;
  end;

  { TFixedSequencePattern }

  TSpellStyle = (Channel, Dontknow, DoDamage, Heal);

  TFixedSequencePattern = class(TSequencePattern)
    Sequence: array of TElement;
    MustClose: boolean;
    Style: TSpellStyle;
    function CanRepeat: boolean;override;
    function MaxLen: Integer;override;
    function Recognize(Game: TGame; Line: TLine): TSpell;override;
    constructor Create(aSeq: string; aStyle: TSpellStyle; aMustClose: boolean = false);
    function CheckLink(Index: Integer; Line: TLine; Node: TSource; FirstNode: TSource; FinalLines, FinalNodes: TObjectList): TLinkCheckResult;override;
  end;

  { TSetSequencePattern }

  TSetSequencePattern = class(TSequencePattern)
    FCanRepeat: Boolean;
    Elements: TElements;
    Style: TSpellStyle;
    function CanRepeat: boolean;override;
    function MaxLen: Integer;override;
    function Recognize(Game: TGame; Line: TLine): TSpell;override;
    constructor CreateOneElem(Elem: TElement; aStyle: TSpellStyle);
    constructor CreateManyElems(Elems: TElements; aStyle: TSpellStyle);
    function CheckLink(Index: Integer; Line: TLine; Node: TSource; FirstNode: TSource; FinalLines, FinalNodes: TObjectList): TLinkCheckResult;override;
  end;

  { THexagram }

  THexagram = class(TSequencePattern)
    function CanRepeat: boolean;override;
    function MaxLen: Integer;override;
    function Recognize(Game: TGame; Line: TLine): TSpell;override;
    function CheckLink(Index: Integer; Line: TLine; Node: TSource; FirstNode: TSource; FinalLines, FinalNodes: TObjectList): TLinkCheckResult;override;
  end;

  { TPentagram }

  TPentagram = class(TSequencePattern)
    function CanRepeat: boolean;override;
    function MaxLen: Integer;override;
    function Recognize(Game: TGame; Line: TLine): TSpell;override;
    function CheckLink(Index: Integer; Line: TLine; Node: TSource; FirstNode: TSource; FinalLines, FinalNodes: TObjectList): TLinkCheckResult;override;
  end;

  { TPlayer }

  TPlayer = class(TUnit)
    CurLine: TLine;
    DrawMode: boolean;
    procedure Draw(C: TCanvas); override;
    procedure Die; override;
    constructor Create(aGame: TGame; aX, aY: extended);override;
    procedure Process; override;
  end;


  { TGame }

  TGame = class
    Objects: TList;
    Player:  TPlayer;
    Scratches: TList;
    Patterns: TList;
    procedure InitPatterns;
    function CheckPatterns(L: TLine): TSpell;
    procedure ProcessTime;
    procedure Clear;
    procedure LoadLevel(Filename: string);
    procedure GenLevel;
    procedure Draw(C: TCanvas);
    function CheckObstracles(Who: TUnit; var X: extended; var Y: extended): boolean;
    constructor Create;
    function NearObject(X, Y: Real; aClass: TGameClass; Exclude: TGameObject = nil): TGameObject;
    function NearestObject(X, Y: Real; aClass: TGameClass; Exclude: TGameObject = nil): TGameObject;
  end;

  const
    ElementColors: TElementColors = (clRed, clBlue, clSkyBlue, clGreen);
    ElementNames: TElementNames = ('F', 'W', 'A', 'E');


function PointsNear(X1, Y1, X2, Y2: Extended): boolean;
function AvgColor(C1, C2: TColor): TColor;
function IntersectLines(X1, Y1, X2, Y2, X3, Y3, X4, Y4: Extended): boolean;
function InTria(X1, Y1, X2, Y2, X3, Y3: Extended; X, Y: Extended): boolean;
function RandomElement: TElement;

implementation

uses uMain;

function PointsNear(X1, Y1, X2, Y2: Extended): boolean;
begin
  Result := Sqr(X1-X2)+Sqr(Y1-Y2)<Sqr(MINDIST);
end;

function AvgColor(C1, C2: TColor): TColor;
var
  R, G, B: Integer;
begin
  C1 := ColorToRGB(C1);
  C2 := ColorToRGB(C2);
  R := (Red(C1)+Red(C2))div 2;
  G := (Green(C1)+Green(C2))div 2;
  B := (Blue(C1)+Blue(C2))div 2;
  Result := RGBToColor(R,G,B);
end;

{Украдено из интернета}
function IntersectLines(X1, Y1, X2, Y2, X3, Y3, X4, Y4: Extended): boolean;

  function Q(ax,ay,bx,by,tx,ty: Extended): Extended;
  begin
    Result := tx*(by-ay)+ty*(ax-bx)+ay*bx-ax*by;
  end;

begin
  Result :=
    (Q(X1,Y1,X2,Y2,X3,Y3)*Q(X1,Y1,X2,Y2,X4,Y4) < 0)and
    (Q(X3,Y3,X4,Y4,X1,Y1)*Q(X3,Y3,X4,Y4,X2,Y2) < 0)
end;

{Украдено из интернета}
function InTria(X1, Y1, X2, Y2, X3, Y3: Extended; X, Y: Extended): boolean;
var
  f1,f2,f3: boolean;
begin
  f1:=(((x-x1)*(y1-y2)+(y-y1)*(x2-x1))>0);
  f2:=(((x-x2)*(y2-y3)+(y-y2)*(x3-x2))>0);
  f3:=(((x-x3)*(y3-y1)+(y-y3)*(x1-x3))>0);
  Result := (f1=f2) and (f2=f3);
end;

function RandomElement: TElement;
begin
  Result := TElement(Random(Ord(High(TElement))+1));
end;

{ TDamageSpell }

class function TDamageSpell.PreciseArea: boolean;
begin
  Result := False;
end;

function TDamageSpell.Callback(U: TUnit; aX, aY: Extended): boolean;
begin
  Result := True;
//  if U.InheritsFrom(TMonster) then U.Hit;
end;

procedure TDamageSpell.Process;
var
  I: integer;
begin
  inherited Process;
  if UseMana then
  begin
    for I := 0 to Game.Objects.Count-1 do
    if TGameObject(Game.Objects[I]).InheritsFrom(TMonster)then
      with TMonster(Game.Objects[I])do
        if MyArea.Inside(X, Y) then Hit;
  end
  else
    Free;
end;

{ THealSpell }

class function THealSpell.PreciseArea: boolean;
begin
  Result := False;
end;

function THealSpell.Callback(U: TUnit; aX, aY: Extended): boolean;
begin
  Result := True;
end;

procedure THealSpell.Process;
begin
  inherited Process;
  with Game.Player do
    if MyArea.Inside(X, Y) then Heal;
  if not UseMana then Free;
end;

{ TPentagram }

function TPentagram.CanRepeat: boolean;
begin
  Result := False;
end;

function TPentagram.MaxLen: Integer;
begin
  Result := 5;
end;

function TPentagram.Recognize(Game: TGame; Line: TLine): TSpell;
var
  I: Integer;
  N: TSource;
  Nodes, Lines: TObjectList;
  CX, CY: Extended;
  L: TLine;
begin
  Nodes := TObjectList.Create;
  Lines := TObjectList.Create;
  try
    N := FullRecognize(Line, Lines, Nodes);
    if (N <> nil) and (N = Line.Finish) then
    begin
      Result := TSummonSpell.Create(Game, TDemon);
      for I := 0 to Length(Lines.Data)-1 do
        Result.AddLine(TLine(Lines.Data[I]));
      TSummonSpell(Result).Activate;
    end
    else
      Result := nil;
  finally
    Nodes.Free;
    Lines.Free;
  end;
end;

function TPentagram.CheckLink(Index: Integer; Line: TLine; Node: TSource;
  FirstNode: TSource; FinalLines, FinalNodes: TObjectList): TLinkCheckResult;

function Inters(I1, I2, J1, J2: Integer): boolean;
begin
  with FinalNodes do Result := IntersectLines(X(I1), Y(I1), X(I2), Y(I2), X(J1), Y(J1), X(J2), Y(J2))
end;

begin
  Result := IsFail;
  {TODO: Инкрементальная проверка пересечений}
  if Index = 4 then
  begin
    if not Inters(1, 2, 3, 4) then exit;
    if not Inters(1, 2, 5, 4) then exit;
    if not Inters(1, 5, 3, 4) then exit;
    if not Inters(1, 5, 3, 2) then exit;
    if not Inters(2, 3, 4, 5) then exit;
  end;
  Result := inherited;
end;

{ TSummonSpell }

constructor TSummonSpell.Create(aGame: TGame; aClass: TUnitClass);
begin
  inherited Create(aGame);
  SummonClass := aClass;
  Summoned := nil;
end;

procedure TSummonSpell.Activate;
var
  aX, aY: Extended;
  I: Integer;
begin
  aX := 0;
  aY := 0;
  for I := 0 to Sources.Count-1 do
    with TSource(Sources[I]) do
    begin
      aX := aX + X;
      aY := aY + Y;
    end;
  aX := aX/Sources.Count;
  aY := aY/Sources.Count;
  Summoned := SummonClass.Create(Game, aX, aY);
end;

procedure TSummonSpell.Process;
begin
  if not UseMana then
  begin
    Summoned.HP := -1;
    Free;
    exit;
  end;
end;

{ TGolem }

procedure TGolem.Draw(C: TCanvas);
begin
  C.Brush.Color := clYellow;
  C.EllipseC(Round(X), Round(Y), 7, 7);
  C.TextOut(Round(X), Round(Y), 'G');
  inherited Draw(C);
end;

constructor TGolem.Create(aGame: TGame; aX, aY: extended);
begin
  inherited;
  Speed := GOLEMSPEED;
  Armor := 1;
end;

procedure TGolem.Process;
var
  U: TUnit;
begin
  U := TUnit(Game.NearestObject(X, Y, TMonster, Self));
  if U <> nil then
    SetDirTo(U.X, U.Y);
  Move;
  U := TUnit(Game.NearObject(X, Y, TMonster, Self));
  if U<>nil then
    U.Hit;
  inherited Process;
end;

{ TDemon }

procedure TDemon.Draw(C: TCanvas);
begin
  C.Brush.Color := clMaroon;
  C.EllipseC(Round(X), Round(Y), 7, 7);
  C.TextOut(Round(X), Round(Y), 'D');
  inherited Draw(C);
end;

constructor TDemon.Create(aGame: TGame; aX, aY: extended);
begin
  inherited;
  Speed := DEMONSPEED;
  Armor := 1;
end;

procedure TDemon.Process;
var
  U: TUnit;
begin
  U := TUnit(Game.NearestObject(X, Y, TUnit, Self));
  if U <> nil then
    SetDirTo(U.X, U.Y);
  Move;
  U := TUnit(Game.NearObject(X, Y, TUnit, Self));
  if U <> nil then
    U.Hit;
  inherited Process;
end;

{ TChannelSpell }

procedure TChannelSpell.Process;
var
  A, B: TSource;
begin
  inherited Process;
  A := TSource(Sources[0]);
  B := TSource(Sources[1]);
  if (A.MP<= MANACOST) and (B.MP <= MANACOST) then
  begin
    Free;
    exit;
  end;
  if A.MP < B.MP then
  begin
    if B.UseMana then
      A.MP := A.MP + MANACOST
    else
      Free;
  end
  else if B.MP < A.MP then
  begin
    if A.UseMana then
      B.MP := B.MP + MANACOST
    else
      Free
  end
end;

constructor TChannelSpell.Create(aGame: TGame; First, Second: TSource);
begin
  inherited Create(aGame);
  AddLine(First.Connected(Second));
  AddLine(First.Connected(Second));
end;

{ TBarrier }

class function TBarrier.PreciseArea: boolean;
begin
  Result := True;
end;

function TBarrier.Callback(U: TUnit; aX, aY: Extended): boolean;
begin
  if U.InheritsFrom(TMonster) and (MyArea.Inside(aX, aY) xor MyArea.Inside(U.X, U.Y)) then
    Result := False
  else
    Result := True;
end;

procedure TBarrier.Process;
begin
  inherited Process;
  if not UseMana then Free;
end;

{ TAreaSpell }

constructor TAreaSpell.Create(aGame: TGame);
begin
  inherited;
  MyArea := TSpecialObstracle.Create(aGame, Self);
end;

procedure TAreaSpell.Activate;
var
  I, J: Integer;
  L: TLine;
  S: TSource;
begin
  if PreciseArea then
  begin
    for I := 0 to Lines.Count - 1 do
    begin
      L := TLine(Lines[I]);
      MyArea.AddPoint(L.Start.X, L.Start.Y);
      for J := Low(L.Points) to High(L.Points) do
        with TExtendedPoint(L.Points[J])do
          MyArea.AddPoint(X, Y);
      MyArea.AddPoint(L.Finish.X, L.Finish.Y);
    end;
  end
  else
  begin
    for I := 0 to Lines.Count - 1 do
    begin
      S := TSource(Sources[I]);
      MyArea.AddPoint(S.X, S.Y);
    end;
  end;
end;

destructor TAreaSpell.Destroy;
begin
  MyArea.Free;
  inherited;
end;

{ TSpecialObstracle }

constructor TSpecialObstracle.Create(aGame: TGame; aCaller: TAreaSpell);
begin
  inherited Create(aGame, 0, 0);
  SetLength(Points, 0);
  Caller := aCaller;
end;

destructor TSpecialObstracle.Destroy;
begin
  SetLength(Points, 0);
  inherited Destroy;
end;

function TSpecialObstracle.Inside(aX, aY: Extended): boolean;
var
  C, I: integer;
begin
  {Процедура проверки принадлежности точки многоугольнику}
  C := 0;
  for I := Low(Points)+1 to High(Points) do
    if IntersectLines(-1000, 0, aX, aY, Points[I-1].X, Points[I-1].Y, Points[I].X, Points[I].Y) then
      Inc(C);
  Result := Odd(C)
end;

procedure TSpecialObstracle.AddPoint(aX, aY: Extended);
begin
  SetLength(Points, Length(Points)+1);
  with Points[High(Points)]do
  begin
    X := aX;
    Y := aY;
  end;
end;

function TSpecialObstracle.CanStep(Who: TUnit; var aX: extended; var aY: extended
  ): boolean;
begin
  Result := Caller.Callback(Who, aX, aY)
end;

procedure TSpecialObstracle.Draw(C: TCanvas);
begin
  {TODO: OpenGL}
  C.Brush.Color := clLime;
  C.TextOut(100,100, '!!!!!!');
end;

procedure TSpecialObstracle.Process;
var
  I: Integer;
  U: TUnit;
begin
  inherited Process;
  for I := Game.Objects.Count-1 downto 0 do
    if TGameObject(Game.Objects[I]).InheritsFrom(TUnit) then
    begin
      U := TUnit(Game.Objects[I]);
      if Inside(U.X, U.Y) then
        Caller.Callback(U, U.X, U.Y);
    end;
end;

{ THexagram }

function THexagram.CanRepeat: boolean;
begin
  Result := False;
end;

function THexagram.MaxLen: Integer;
begin
  Result := 6;
end;

function THexagram.Recognize(Game: TGame; Line: TLine): TSpell;
var
  I, J: Integer;
  N: TSource;
  Nodes, Lines: TObjectList;
  CX, CY: Extended;
  L: TLine;
begin
  Nodes := TObjectList.Create;
  Lines := TObjectList.Create;
  try
    N := FullRecognize(Line, Lines, Nodes);
    if (N <> nil) and (N = Line.Finish) then
    begin
      Result := TSummonSpell.Create(Game, TGolem);
      for I := 0 to Length(Lines.Data)-1 do
        Result.AddLine(TLine(Lines.Data[I]));
      for I := 0 to Length(Nodes.Data)-1 do
      for J := 0 to Length(Nodes.Data)-1 do
      begin
        L := TSource(Nodes.Data[I]).Connected(TSource(Nodes.Data[J]));
        if L <> nil then
          Result.AddLine(L);
      end;
      TSummonSpell(Result).Activate;
    end
    else
      Result := nil;
  finally
    Nodes.Free;
    Lines.Free;
  end;
end;

function THexagram.CheckLink(Index: Integer; Line: TLine; Node: TSource;
  FirstNode: TSource; FinalLines, FinalNodes: TObjectList): TLinkCheckResult;
var
  CX, CY: Extended;
begin
  Result := IsFail;
  if (Index > 0) and (Node.Connected(TSource(FinalNodes.PrevNode(2))) = nil)then
    exit
  else if Index = 5 then
    with FinalNodes do
    begin
      if InTria(X(2), Y(2), X(4), Y(4), X(6), Y(6), X(1), Y(1)) then exit;
      if InTria(X(2), Y(2), X(4), Y(4), X(6), Y(6), X(3), Y(3)) then exit;
      if InTria(X(2), Y(2), X(4), Y(4), X(6), Y(6), X(5), Y(5)) then exit;

      if InTria(X(1), Y(1), X(3), Y(3), X(5), Y(5), X(2), Y(2)) then exit;
      if InTria(X(1), Y(1), X(3), Y(3), X(5), Y(5), X(4), Y(4)) then exit;
      if InTria(X(1), Y(1), X(3), Y(3), X(5), Y(5), X(6), Y(6)) then exit;

      CX := (X(1)+X(3)+X(5))/3;
      CY := (Y(1)+Y(3)+Y(5))/3;
      if not InTria(X(2), Y(2), X(4), Y(4), X(6), Y(6), CX, CY) then exit;
      CX := (X(2)+X(4)+X(6))/3;
      CY := (Y(2)+Y(4)+Y(6))/3;
      if not InTria(X(1), Y(1), X(3), Y(3), X(5), Y(5), CX, CY) then exit;
    end;
  Result := inherited;
end;

{ TSetSequencePattern }

function TSetSequencePattern.CanRepeat: boolean;
begin
  Result := FCanRepeat;
end;

function TSetSequencePattern.MaxLen: Integer;
begin
  Result := 10000;
end;

function TSetSequencePattern.Recognize(Game: TGame; Line: TLine): TSpell;
var
  I: Integer;
  N: TSource;
  Nodes, Lines: TObjectList;
begin
  Nodes := TObjectList.Create;
  Lines := TObjectList.Create;
  N := FullRecognize(Line, Lines, Nodes);
  if (N <> nil) and (N = Line.Finish)then
  begin
    case Style of
    DoDamage: Result := TDamageSpell.Create(Game);
    Heal: Result := THealSpell.Create(Game);
    end;
    for I := 0 to Length(Lines.Data)-1 do
      Result.AddLine(TLine(Lines.Data[I]));
    TAreaSpell(Result).Activate;
  end
  else
    Result := nil;
  Nodes.Free;
  Lines.Free;
  log('Recog done!');
end;

constructor TSetSequencePattern.CreateOneElem(Elem: TElement; aStyle: TSpellStyle);
begin
  inherited Create;
  Style := aStyle;
  FCanRepeat := True;
  Elements := [];
  Include(Elements, Elem);
end;

constructor TSetSequencePattern.CreateManyElems(Elems: TElements; aStyle: TSpellStyle);
begin
  inherited Create;
  Style := aStyle;
  FCanRepeat := False;
  Elements := Elems;
end;

function TSetSequencePattern.CheckLink(Index: Integer; Line: TLine;
  Node: TSource; FirstNode: TSource; FinalLines, FinalNodes: TObjectList): TLinkCheckResult;
begin
  if Node.Element in Elements then
  begin
    Result := inherited;
    if (Result = PartialOK) and (Node = FirstNode) then
    begin
      if Index > 1 then
        Result := LastOK
      else
        Result := IsFail;
    end;
  end
  else
    Result := IsFail;
end;

{ TObjectList }

constructor TObjectList.Create;
begin
  inherited;
  SetLength(Data, 0);
end;

destructor TObjectList.Destroy;
begin
  SetLength(Data, 0);
  inherited Destroy;
end;

procedure TObjectList.PushObject(S: TObject);
begin
  SetLength(Data, Length(Data)+1);
  Data[High(Data)] := S;
end;

procedure TObjectList.PopObject;
begin
  SetLength(Data, Length(Data)-1);
end;

function TObjectList.FindSource(S: TObject): boolean;
var
  I: Integer;
begin
  for I := Low(Data) to High(Data) do
    if Data[I] = S then
    begin
      Result := True;
      exit
    end;
  Result := False
end;

function TObjectList.PrevNode(N: Integer): TObject;
begin
  if N > Length(Data) then
    Result := nil
  else
    Result := Data[High(Data)-N+1];
end;

function TObjectList.X(I: Integer): Extended;
begin
  Result := TSource(PrevNode(I)).X;
end;

function TObjectList.Y(I: Integer): Extended;
begin
  Result := TSource(PrevNode(I)).Y;
end;

{ TFixedSequencePattern }

function TFixedSequencePattern.CanRepeat: boolean;
begin
  Result := True;
end;

function TFixedSequencePattern.MaxLen: Integer;
begin
  Result := Length(Sequence);
end;

function TFixedSequencePattern.Recognize(Game: TGame; Line: TLine): TSpell;
var
  I: Integer;
  N: TSource;
  Nodes, Lines: TObjectList;
begin
  Nodes := TObjectList.Create;
  Lines := TObjectList.Create;
  N := FullRecognize(Line, Lines, Nodes);
  if (N <> nil) and (not MustClose or (N = Line.Finish))then
  begin
    case Style of
    Channel:
      begin
       Result := TChannelSpell.Create(Game, Line.Start, Line.Finish);
      end;
    Dontknow: ;
    end;
  end
  else
    Result := nil;
  Nodes.Free;
  Lines.Free;
  log('Recog done!');
end;

constructor TFixedSequencePattern.Create(aSeq: string; aStyle: TSpellStyle; aMustClose: boolean);
var
  I, N: integer;
begin
  inherited Create;
  N := Length(aSeq);
  Style := aStyle;
  MustClose := aMustClose;
  SetLength(Sequence, N);
  for I := 0 to N-1 do
    case aSeq[I+1] of
      'F','f': Sequence[I] := Fire;
      'W','w': Sequence[I] := Water;
      'A','a': Sequence[I] := Air;
      'E','e': Sequence[I] := Earth;
    end;
end;

function TFixedSequencePattern.CheckLink(Index: Integer; Line: TLine; Node: TSource;
  FirstNode: TSource; FinalLines, FinalNodes: TObjectList): TLinkCheckResult;
begin
  Log(Format('C: %s, need %s', [ElementNames[Node.Element], ElementNames[Sequence[Index]]]));
  if Node.Element <> Sequence[Index] then
    Result := IsFail
  else
    Result := inherited;
end;

{ TSequencePattern }

function TSequencePattern.CanRepeat: boolean;
begin
  Result := False
end;

function TSequencePattern.RecognizePart(CurIndex: Integer; FromNode: TSource;
  FirstNode: TSource; Exclude, FinalLines, FinalNodes: TObjectList): TSource;
var
  I: Integer;
  L: TLine;
  Next: TSource;
begin
  {MAIN recursive procedure}
  Result := nil;
  Log(Format('Step%d: Node=%s', [CurIndex, FromNode.Name]));
  log('L: '+IntToStr(Length(FinalNodes.Data)));
  for I := 0 to Length(FinalNodes.Data)-1 do
    Log(IntToStr(I)+': '+TSource(FinalNodes.Data[I]).Name);
  for I := 0 to FromNode.Lines.Count - 1 do
  begin
    L := TLine(FromNode.Lines[I]);
    Next := L.OtherNode(FromNode);
    if Exclude.FindSource(Next) then
    begin
      log(Format('%s Excluded!', [Next.Name]));
      Result := nil;
      continue;
    end;
    Log(Format('Checking link from %s to %s', [FromNode.Name, Next.Name]));
    case CheckLink(CurIndex, L, Next, FirstNode, FinalLines, FinalNodes) of
    PartialOK:
      begin
        Log('PartOK');
        if Next <> FirstNode then
        begin
          Exclude.PushObject(FromNode);
          FinalNodes.PushObject(Next);
          FinalLines.PushObject(L);
          Result := RecognizePart(CurIndex+1, Next, FirstNode, Exclude, FinalLines, FinalNodes);
          Exclude.PopObject;
          if Result <> nil then
            exit
          else
          begin
            FinalNodes.PopObject;
            FinalLines.PopObject;
          end
        end
      end;
    FullOK:
      begin
        Log('FullOK');
        FinalNodes.PushObject(Next);
        FinalLines.PushObject(L);
        if Next <> FirstNode then
        begin
          Exclude.PushObject(FromNode);
          Result := RecognizePart(CurIndex+1, Next, FirstNode, Exclude, FinalLines, FinalNodes);
          Exclude.PopObject;
          if Result = nil then
          begin
            FinalNodes.PopObject;
            FinalLines.PopObject;
          end
        end;
        if Result = nil then
          Result := Next;
        exit
      end;
    IsFail:
      begin
        Log('Fail');
        continue;
      end;
    LastOK:
      begin
        FinalNodes.PushObject(Next);
        FinalLines.PushObject(L);
        Log('LastOK');
        Result := Next;
        exit
      end;
    end;
  end;
  Result := nil;
end;

function TSequencePattern.CheckLink(Index: Integer; Line: TLine; Node: TSource;
  FirstNode: TSource; FinalLines, FinalNodes: TObjectList): TLinkCheckResult;
begin
  if (not CanRepeat) and (Line.Start.Element = Line.Finish.Element) then
    Result := IsFail
  else if Index >= MaxLen-1 then
    Result := LastOK
  else
    Result := PartialOK;
end;

function TSequencePattern.FullRecognize(aLine: TLine; FinalLines, FinalNodes: TObjectList): TSource;
var
  L: TObjectList;
begin
  L := TObjectList.Create;
  {Хак: уберем первую линию}
  aLine.Start.Lines.Remove(aLine);
  FinalLines.PushObject(aLine);
  FinalNodes.PushObject(aLine.Finish);
  FinalNodes.PushObject(aLine.Start);

  case CheckLink(0, aLine, aLine.Start, aLine.Finish, FinalLines, FinalNodes) of
  PartialOK:
    begin
      Result := RecognizePart(1, aLine.Start, aLine.Finish, L, FinalLines, FinalNodes);
    end;
  FullOK:
    begin
      Result := RecognizePart(1, aLine.Start, aLine.Finish, L, FinalLines, FinalNodes);
      if Result = nil then Result := aLine.Start;
    end;
  IsFail:
    Result := nil;
  LastOK:
    Result := aLine.Start;
  end;
  {Хак: вернем первую линию}
  aLine.Start.Lines.Add(aLine);
  L.Free;
end;

{ TDemoSpell }

procedure TDemoSpell.Process;
begin

end;

constructor TDemoSpell.Create(aGame: TGame);
begin
  inherited Create(aGame);
end;

{ TSpell }

constructor TSpell.Create(aGame: TGame);
begin
  inherited Create(aGame, 0,0);
  Sources := TList.Create;
  Lines := TList.Create;
end;

procedure TSpell.AddLine(L: TLine);
begin
  L.Start.Lines.Remove(L);
  L.Finish.Lines.Remove(L);
  Lines.Add(L);
  if Sources.IndexOf(L.Start) < 0 then
    Sources.Add(L.Start);
  if Sources.IndexOf(L.Finish) < 0 then
    Sources.Add(L.Finish);
end;

procedure TSpell.Draw(C: TCanvas);
var
  I: Integer;
begin
  for I := 0 to Lines.Count-1 do
    TLine(Lines[I]).Draw(C, Active);
end;

procedure TSpell.Process;
begin
  inherited Process;
end;

function TSpell.UseMana: boolean;
var
  I: Integer;
begin
  Result := True;
  for I := Sources.Count-1 downto 0 do
    if not TSource(Sources[I]).UseMana then
    begin
      Result := False;
    end;
end;

destructor TSpell.Destroy;
var
  I: Integer;
begin
  Sources.Free;
  for I := 0 to Lines.Count-1 do
    Game.Scratches.Add(Lines[I]);
  Lines.Free;
  inherited Destroy;
end;

{ TCirclePattern }

function TCirclePattern.Recognize(Game: TGame; Line: TLine): TSpell;
begin
  if Line.Start = Line.Finish then
  begin
    Result := TBarrier.Create(Game);
    Result.AddLine(Line);
    TBarrier(Result).Activate;
  end
  else
    Result := nil
end;

{ TLine }

constructor TLine.Create(aGame: TGame; aStart: TSource);
begin
  inherited Create;
  Game := aGame;
  Start := aStart;
  Finish := nil;
  SetLength(Points, 0);
end;

procedure TLine.AddPoint(aX, aY: Extended; Angle: Real);
begin
  if Angle = LastAngle then exit;
  if PointsNear(aX, aY, Start.X, Start.Y) then exit;
  LastAngle := Angle;
  SetLength(Points, Length(Points)+1);
  with Points[High(Points)]do
  begin
    X := aX;
    Y := aY;
  end;
end;

function TLine.Finalize(aFinish: TSource): boolean;
begin
  log('Finalize');
  Result := (Start <> aFinish) or (Length(Points) > 0);
  if not Result then exit;
  Finish := aFinish;
  if Finish = nil then
    Game.Scratches.Add(Self)
  else
  begin
    Start.Lines.Add(Self);
    Finish.Lines.Add(Self);
    log('Checking line...');
    Game.CheckPatterns(Self);
  end;
  log('Fin done');
end;

procedure TLine.Draw(C: TCanvas; Style: TLineStyle);
var
  I: Integer;
begin
  case Style of
  Scratch:
    begin
      C.Pen.Color := clDkGray;
    end;
  Current:
    begin
      C.Pen.Color := clWhite;
    end;
  Ready:
    begin
      C.Pen.Color := AvgColor(ElementColors[Start.Element], ElementColors[Finish.Element]);
    end;
  Active:
    begin
      C.Pen.Color := AvgColor(ElementColors[Start.Element], ElementColors[Finish.Element]);
      C.Pen.Width := 3;
    end;
  end;
  C.MoveTo(Round(Start.X), Round(Start.Y));
  for I := Low(Points) to High(Points) do
    with Points[I]do
      C.LineTo(Round(X), Round(Y));
  if Finish <> nil then
    C.LineTo(Round(Finish.X), Round(Finish.Y))
  else if Self = Game.Player.CurLine then
    C.LineTo(Round(Game.Player.X), Round(Game.Player.Y));
  C.Pen.Width := 1;
end;

function TLine.OtherNode(S: TSource): TSource;
begin
  if S = Start then
    Result := Finish
  else
    Result := Start
end;

{ TSource }

procedure TSource.Draw(C: TCanvas);
var
  bx, by, d: Integer;
  I: Integer;
  L: TLine;
begin
  C.Pen.Color := clBlack;
  bx := Round(X);
  by := Round(Y);
  d := 5;
  C.Brush.Color := clDkGray;
  C.Polygon([Point(bx-d, by), Point(bx, by-d), Point(bx+d, by), Point(bx, by+d)]);
  d := Round(5*MP);
  C.Brush.Color := ElementColors[Element];
  C.Polygon([Point(bx-d, by), Point(bx, by-d), Point(bx+d, by), Point(bx, by+d)]);
  C.TextOut(bx, by, Name+':'+IntToStr(Round(MP*100))+'%');
  for I := 0 to Lines.Count-1 do
  begin
    L := TLine(Lines[I]);
    if L.Finish = self then L.Draw(C, Ready);
  end;
end;

function TSource.UseMana: boolean;
begin
  MP := MP - MANACOST;
  Result := MP >= 0;
end;

constructor TSource.Create(aGame: TGame; ax, ay, aMP: extended;
  aElement: TElement);
begin
  inherited Create(aGame, aX, aY);
  MP := aMP;
  Element := aElement;
  Lines := TList.Create;
end;

destructor TSource.Destroy;
var
  I: Integer;
begin
  if (Game.Player.CurLine <> nil) and (Game.Player.CurLine.Start = Self) then
  begin
    log('Destroy!');
    Game.Player.CurLine.Finalize(nil);
    Game.Player.CurLine := nil
  end;
  for I := 0 to Lines.Count-1 do
    Game.Scratches.Add(Lines[I]);
  Lines.Free;
  inherited;
end;

procedure TSource.Process;
begin
  inherited Process;
  if MP <= 0 then
  begin
    Game.Objects.Remove(Self);
//    Free;{TODO; memory leak}
  end;
end;

function TSource.Connected(Node: TSource): TLine;
var
  I: Integer;
begin
  for I := 0 to Lines.Count-1 do
    if TLine(Lines[I]).OtherNode(Self) = Node then
    begin
      Result := TLine(Lines[I]);
      exit
    end;
  Result := nil
end;

{ TGameObject }

destructor TGameObject.Destroy;
begin
  log('Destroy');
  Broken := True;
  Game.Objects.Remove(Self);
  inherited Destroy;
end;

constructor TGameObject.Create(aGame: TGame; aX, aY: extended);
begin
  inherited Create;
  Broken := False;
  Game := aGame;
  Game.Objects.Add(Self);
  X := aX;
  Y := aY;
end;

procedure TGameObject.Process;
begin

end;

{ TMonster }

procedure TMonster.Draw(C: TCanvas);
begin
  C.Brush.Color := clRed;
  C.EllipseC(Round(X), Round(Y), 5, 5);
  C.TextOut(Round(X), Round(Y), 'Z');
  inherited;
end;

constructor TMonster.Create(aGame: TGame; aX, aY: extended);
begin
  inherited;
  Speed := MONSTERSPEED;
  Armor := 0;
end;

procedure TMonster.Process;
begin
  SetDirTo(Game.Player.X, Game.Player.Y);
  Move;
  if PointsNear(X, Y, Game.Player.X, Game.Player.Y) then
    Game.Player.Hit;
  log('processed');
  inherited;
  log('processed ok');
end;

{ TPlayer }

procedure TPlayer.Draw(C: TCanvas);
begin
  C.Pen.Color   := clPurple;
//  C.Brush.Color := clPurple;
//  C.EllipseC(Round(X), Round(Y), 5, 5);
  C.Brush.Color := clNone;
  C.TextOut(Round(X), Round(Y), '@');
  C.Pen.Color := clLime;
  C.Line(Round(X), Round(Y), Round(TargetX), Round(TargetY));
  if CurLine <> nil then
    CurLine.Draw(C, Current);
  inherited;
end;

procedure TPlayer.Die;
begin
  ShowMessage('Game OVER');
  Application.Terminate;
  inherited Destroy;
end;

constructor TPlayer.Create(aGame: TGame; aX, aY: extended);
begin
  inherited;
  Speed := PLAYERSPEED;
  Armor := PLAYERARMOR;
  DrawMode := False;
  CurLine := nil;
end;

procedure TPlayer.Process;
var
  N: TSource;
begin
  N := TSource(Game.NearObject(X,Y,TSource));
  if CurLine <> nil then
  begin
    if N <> nil then
    begin
      log('Start fin');
      if CurLine.Finalize(N) then
      begin
        log('fin is ok');
        CurLine := nil;
        CurLine := TLine.Create(Game, N);
      end
    end
    else if not DrawMode then
    begin
      CurLine.AddPoint(X, Y, -1000);
      CurLine.Finalize(nil);
      CurLine := nil
    end
    else
      CurLine.AddPoint(X, Y, arctan2(TargetY-Y, TargetX-X));
  end
  else if DrawMode and (N <> nil) then
  begin
    CurLine := TLine.Create(Game, N);
  end;
  Move;
  inherited Process;
end;

{ TRectObstracle }

constructor TRectObstracle.Create(aGame: TGame; ax1, ay1, ax2, ay2: extended);
begin
  inherited Create(aGame, ax1, ay1);
  X2 := ax2;
  Y2 := ay2;
end;

procedure TRectObstracle.Draw(C: TCanvas);
begin
  C.Brush.Color := clGray;
  C.FillRect(Round(X), Round(Y), Round(X2), Round(Y2));
end;

function TRectObstracle.CanStep(Who: TUnit; var aX: extended; var aY: extended): boolean;
begin
  Result := not (InRange(aX, X, X2) and InRange(aY, Y, Y2));
end;

{ TGame }

procedure TGame.InitPatterns;
begin
  //Barrier
  Patterns.Add(TCirclePattern.Create);
  //Channel

  Patterns.Add(TFixedSequencePattern.Create('FF', Channel, True));
  Patterns.Add(TFixedSequencePattern.Create('EE', Channel, True));
  Patterns.Add(TFixedSequencePattern.Create('WW', Channel, True));
  Patterns.Add(TFixedSequencePattern.Create('AA', Channel, True));
  //Don't know
{  Patterns.Add(TFixedSequencePattern.Create('FAWE', Dontknow, True));
  Patterns.Add(TFixedSequencePattern.Create('FEWA', Dontknow, True));
  Patterns.Add(TFixedSequencePattern.Create('WAFE', Dontknow, True));
  Patterns.Add(TFixedSequencePattern.Create('WEFA', Dontknow, True));
  Patterns.Add(TFixedSequencePattern.Create('AFEW', Dontknow, True));
  Patterns.Add(TFixedSequencePattern.Create('AWEF', Dontknow, True));
  Patterns.Add(TFixedSequencePattern.Create('EWAF', Dontknow, True));
  Patterns.Add(TFixedSequencePattern.Create('EFAW', Dontknow, True));}

  //Heal
  Patterns.Add(TSetSequencePattern.CreateManyElems([Earth, Water, Air], Heal));
  //Damage
  Patterns.Add(TSetSequencePattern.CreateOneElem(Fire, DoDamage));

  //Summon
  Patterns.Add(THexagram.Create);
  Patterns.Add(TPentagram.Create);

end;

function TGame.CheckPatterns(L: TLine): TSpell;
var
  I: Integer;
  P: TPattern;
  E: TElement;
  Sp: TSpell;
begin
  Result := nil;
  for I := 0 to Patterns.Count-1 do
  begin
    Result := TPattern(Patterns[I]).Recognize(Self, L);
    if Result <> nil then exit;
  end;
end;

procedure TGame.ProcessTime;
var
  I: integer;
  O: TGameObject;
  X, Y: Extended;
begin
  I := 0;
  for I := Objects.Count - 1 downto 0 do
  begin
    if I < Objects.Count then
    begin
      O := TGameObject(Objects[I]);
      O.Process;
    end;
  end;
  if random < SPAWNRATE then
  begin
    case random(3) of
      0: begin X := 0; Y := Random*MAXY; end;
      1: begin X := MAXX; Y := Random*MAXY; end;
      2: begin X := Random*MAXX; Y := 0; end;
      3: begin X := Random*MAXX; Y := MAXY; end;
    end;
    TMonster.Create(Self, X, Y);
  end;
end;

procedure TGame.Clear;
var
  I: integer;
begin
  for I := Objects.Count - 1 downto 0 do
    TGameObject(Objects[I]).Free;
  Player := nil;
end;

procedure TGame.LoadLevel(Filename: string);
var
  ff: TextFile;
  X, Y, X2, Y2, I, N, Spd: integer;
  U:  TUnit;
  O:  TObstracle;
begin
  Clear;
  AssignFile(ff, Filename);
  Reset(ff);
  ReadLn(ff, X, Y);
  Player := TPlayer.Create(Self, X, Y);
  ReadLn(ff, N);
  for I := 1 to N do
  begin
    ReadLn(ff, X, Y);
    TDemon.Create(Self, X, Y);
  end;
  ReadLn(ff, N);
  for I := 1 to N do
  begin
    ReadLn(ff, X, Y);
    TMonster.Create(Self, X, Y);
  end;
  ReadLn(ff, N);
  for I := 1 to N do
  begin
    ReadLn(ff, X, Y, X2, Y2);
    TRectObstracle.Create(Self, X, Y, X2, Y2);
  end;
  CloseFile(ff);
end;

procedure TGame.GenLevel;
var
  I: Integer;
begin
  Clear;
  Player := TPlayer.Create(Self, MAXX/2, MAXY/2);
  for I := 0 to 20 do
    TSource.Create(Self, Random*MAXX, Random*MAXY, 1, RandomElement);
end;

procedure TGame.Draw(C: TCanvas);
var
  I: integer;
begin
  C.Brush.Color := clBlack;
  C.Pen.Color   := clRed;
  C.Rectangle(0, 0, MAXX, MAXY);
  for I := 0 to Scratches.Count-1 do
    TLine(Scratches[I]).Draw(C, Scratch);
  for I := Objects.Count - 1 downto 0 do
    TGameObject(Objects[I]).Draw(C);
end;

function TGame.CheckObstracles(Who: TUnit; var X: extended; var Y: extended): boolean;
var
  I: integer;
begin
  {TODO: Обрабатывать смерть юнита}
  Result := False;
  if not (InRange(X, 0, MAXX) and InRange(Y, 0, MAXY)) then
    exit;
  for I := 0 to Objects.Count - 1 do
  begin
    if TGameObject(Objects[I]).InheritsFrom(TObstracle) and not
      TObstracle(Objects[I]).CanStep(Who, X, Y) then
      exit;
  end;
  Result := True;
end;

constructor TGame.Create;
begin
  inherited;
  Objects := TList.Create;
  Scratches := TList.Create;
  Patterns := TList.Create;
  Player  := TPlayer.Create(Self, 100, 100);
  InitPatterns;
end;

function TGame.NearObject(X, Y: Real; aClass: TGameClass; Exclude: TGameObject = nil): TGameObject;
var
  I: Integer;
  O: TGameObject;
begin
  for I := 0 to Objects.Count - 1 do
  begin
    O := TGameObject(Objects[I]);
    if O.InheritsFrom(aClass) and (O <> Exclude) and
      PointsNear(X, Y, O.X, O.Y) then
    begin
      Result := O;
      exit
    end;
  end;
  Result := nil;
end;

function TGame.NearestObject(X, Y: Real; aClass: TGameClass;
  Exclude: TGameObject): TGameObject;
var
  I: Integer;
  O: TGameObject;
  MinRange, D: Extended;
begin
  Result := nil;
  MinRange := Sqr(MAXX + MAXY);
  for I := 0 to Objects.Count - 1 do
  begin
    O := TGameObject(Objects[I]);
    if (not O.InheritsFrom(aClass)) or (O = Exclude)then continue;
    D := Sqr(X-O.X)+Sqr(Y-O.Y);
    if D < MinRange then
    begin
      MinRange := D;
      Result := O;
    end;
  end;
end;

{ TUnit }

procedure TUnit.Move;
var
  R, Angle, aX, aY: extended;
begin
  R := sqr(X - TargetX) + sqr(Y - TargetY);
  if R < sqr(Speed) then
  begin
    aX := TargetX;
    aY := TargetY;
  end
  else
  begin
    Angle := arctan2(TargetY - Y, TargetX - X);
    aX    := X + Speed * cos(Angle);
    aY    := Y + Speed * sin(Angle);
  end;
  if Game.CheckObstracles(Self, aX, aY) then
  begin
    X := aX;
    Y := aY;
  end;
end;

procedure TUnit.Hit;
begin
  if random < armor then exit;
  HP := HP - random * DAMAGE;
end;

procedure TUnit.Die;
begin
  Destroy;
end;

procedure TUnit.SetDirTo(ToX, ToY: extended);
begin
  TargetX := ToX;
  TargetY := ToY;
end;

procedure TUnit.Draw(C: TCanvas);
var
  x1, x2, xr, y1, y2: integer;
begin
  X1 := round(X) - 10;
  X2 := round(X) + 10;
  Y1 := round(Y) - 15;
  Y2 := round(Y) - 10;
  XR := round((X2 - X1) * HP) + X1;
  C.Pen.Color := clBlack;
  C.Brush.Color := clRed;
  C.Rectangle(XR, Y1, X2, Y2);
  C.Brush.Color := clLime;
  C.Rectangle(X1, Y1, XR, Y2);
end;

procedure TUnit.Process;
begin
  inherited;
  if HP <= 0 then
    Die
end;

constructor TUnit.Create(aGame: TGame; aX, aY: extended);
begin
  inherited;
  TargetX := aX;
  TargetY := aY;
  HP := 1;
end;

procedure TUnit.Heal;
begin
  HP := min(1, HP+HEALPOWER)
end;

end.

